/*
 *  Copyright (c) 2007 Jiri Benc <jbenc@upir.cz>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <string.h>
#include <stdio.h>
#include "chordlib.h"

/*
 * C  C# D  D# E  F  F# G  G# A  Bb H
 * 0  1  2  3  4  5  6  7  8  9  10 11
 */

signed char strings[STRINGS_CNT] = { 9, 4, 0, 7 };

#define TONES_MAX	4
static signed char chorddef[][TONES_MAX] = {
	{ 0, 4, 7, -1 },	// dur
	{ 0, 3, 7, -1 },	// mi
	{ 0, 4, 7, 10 },	// 7
	{ 0, 3, 7, 10 },	// mi7
	{ 0, 4, 7, 11 },	// 7maj
	{ 0, 3, 6, 9 },		// dim
	{ 0, 5, 7, -1 },	// sus4
	{ 0, 2, 7, -1 },	// sus2
	{ 0, 4, 8, -1 },	// (5)+
	{ 0, 4, 6, -1 },	// 5-
	{ 0, 3, 6, -1 },	// mi5-
	{ 0, 4, 7, 9 },		// (add)6
	{ 0, 3, 7, 9 },		// mi6
	{ 0, 4, 8, 10 },	// 7/5+
	{ 0, 4, 6, 10 },	// 7/5-
	{ 0, 3, 6, 10 },	// mi7/5-
	{ 0, 3, 7, 11 },	// mi7maj
	{ 0, 4, 7, 2 },		// (add)9
	{ 0, 3, 7, 2 },		// mi9
};
#define CHORDDEF_CNT	(sizeof(chorddef) / sizeof(chorddef[0]))

char *(chordsuff_all[][CHORDDEF_CNT]) = {
	{
	"maj",
	"m",
	"7",
	"m7",
	"7maj",
	"dim",
	"sus4",
	"sus2",
	"+",
	"5-",
	"mi5-",
	"6",
	"m6",
	"7/5+",
	"7/5-",
	"m7/5-",
	"m7maj",
	"9",
	"m9",
	},
	{
	"dur",
	"m",
	"7",
	"m7",
	"7maj",
	"dim",
	"sus4",
	"sus2",
	"(5)+",
	"5-",
	"mi5-",
	"(add)6",
	"m6",
	"7/5+",
	"7/5-",
	"m7/5-",
	"m7maj",
	"(add)9",
	"m9",
	},
};
char **chordsuff;

#define CHORDNAMES_CNT	12
static char *(chordnames_all[][CHORDNAMES_CNT]) = {
	{ "C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B" },
	{ "C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "Bb", "H" },
};
char **chordnames;

char *(langnames[LANG_CNT]) = {
	"English", "Czech",
};

static signed char tones[TONES_MAX];
static int tones_usage[TONES_MAX];
static int tones_cnt;

#define CHORDS_MAX	128
static unsigned chords[CHORDS_MAX];
static signed char chords_bar[CHORDS_MAX];
static int chords_cnt;
/* unsigned is not enough for more than 4 strings */

static signed char cur_chord[STRINGS_CNT];

#define get_pos(chord, i)	(((chord) >> ((i) * 8)) & 0xff)
#define min(a, b)		((a) < (b) ? (a) : (b))

void set_chord_lang(int lang)
{
	if (lang < 0 || lang >= LANG_CNT)
		lang = 0;
	chordnames = chordnames_all[lang];
	chordsuff = chordsuff_all[lang];
}

void set_tuning(signed char tuning[])
{
	int i;

	for (i = 0; i < STRINGS_CNT; i++)
		strings[i] = tuning[i];
}

static void add_cur_chord(void)
{
	int i;
	unsigned chord;
	signed char max;

	for (i = 0; i < tones_cnt; i++) {
		if (tones_usage[i] == 0)
			return;
	}
	if (chords_cnt == CHORDS_MAX) {
		fprintf(stderr, "Max chords reached\n");
		return;
	}
	chord = 0;
	max = 0;
	for (i = 0; i < STRINGS_CNT; i++) {
		chord |= (unsigned)cur_chord[i] << (i * 8);
		if (cur_chord[i] > max)
			max = cur_chord[i];
	}
	chords[chords_cnt] = chord;
	chords_bar[chords_cnt] = max;
	chords_cnt++;
}

static void place_tone(int start_bar, int cur_string);

static void place_one_tone(int start_bar, int cur_string, int bar)
{
	int i;

	for (i = 0; i < tones_cnt; i++)
		if (tones[i] == ((strings[cur_string] + bar) % 12)) {
			tones_usage[i]++;
			cur_chord[cur_string] = bar;
			place_tone(start_bar, cur_string + 1);
			tones_usage[i]--;
			break;
		}
}

static void place_tone(int start_bar, int cur_string)
{
	int i;

	if (cur_string == STRINGS_CNT) {
		add_cur_chord();
		return;
	}
	place_one_tone(start_bar, cur_string, 0);
	for (i = start_bar; i < min(start_bar + BAR_RANGE, LAST_BAR); i++)
		place_one_tone(start_bar, cur_string, i);
}

static void find_chords(void)
{
	int i;

	memset(tones_usage, 0, sizeof(tones_usage));
	chords_cnt = 0;
	for (i = 1; i < LAST_BAR - BAR_RANGE + 1; i++)
		place_tone(i, 0);
}

#define compare_bars(a, b)	((a) < (b) ? -1 : ((a) == (b) ? 0 : 1))

/* this doesn't reflex the number of strings - the reason is a speed */
#define chord_cmp_token(ch)	((unsigned)((ch) & 0xff) + (((ch) >> 8) & 0xff) + \
				 (((ch) >> 16) & 0xff) + (((ch) >> 24) & 0xff))
static int compare_chords(unsigned a, unsigned b)
{
	unsigned ta, tb;
	int i;

	ta = chord_cmp_token(a);
	tb = chord_cmp_token(b);
	if (ta < tb)
		return -1;
	if (ta > tb)
		return 1;
	for (i = 0; i < STRINGS_CNT; i++) {
		ta = get_pos(a, i);
		tb = get_pos(b, i);
		if (ta < tb)
			return -1;
		if (ta > tb)
			return 1;
	}
	return 0;
}

static void sort_chords(void)
{
	int i, j, cnt, comp;
	unsigned tmp_u;
	signed char tmp_ch;

	for (i = chords_cnt; i > 0; i--) {
		cnt = 0;
		for (j = 0; j < i - 1; j++) {
			comp = compare_bars(chords_bar[j], chords_bar[j + 1]);
			if (!comp)
				comp = compare_chords(chords[j], chords[j + 1]);
			if (comp > 0) {
				tmp_u = chords[j];
				chords[j] = chords[j + 1];
				chords[j + 1] = tmp_u;
				tmp_ch = chords_bar[j];
				chords_bar[j] = chords_bar[j + 1];
				chords_bar[j + 1] = tmp_ch;
				cnt++;
			}
		}
		if (!cnt)
			break;
	}
}

static void uniq_chords(void)
{
	int i, last;

	if (chords_cnt == 0)
		return;
	last = 0;
	for (i = 1; i < chords_cnt; i++) {
		if (chords[last] != chords[i]) {
			chords[++last] = chords[i];
			chords_bar[last] = chords_bar[i];
		}
	}
	chords_cnt = last + 1;
}

int get_chordnames_cnt(void)
{
	return CHORDNAMES_CNT;
}

int get_chordsuff_cnt(void)
{
	return CHORDDEF_CNT;
}

char *get_tone_name(int string_idx, unsigned chord)
{
	return chordnames[(get_pos(chord, string_idx) + strings[string_idx]) % 12];
}

void generate_chords(int name, int suff)
{
	int i;

	tones_cnt = 0;
	for (i = 0; i < TONES_MAX; i++) {
		if (chorddef[suff][i] < 0)
			break;
		tones[i] = (chorddef[suff][i] + name) % 12;
		tones_cnt++;
	}
	find_chords();
	sort_chords();
	uniq_chords();
}

int get_chords_cnt(void)
{
	return chords_cnt;
}

unsigned get_chord(int idx)
{
	return chords[idx];
}

int get_chord_start_bar(int idx)
{
	return chords_bar[idx] < BAR_RANGE ? 0 : chords_bar[idx] - BAR_RANGE;
}

static int basic_key_chords[BASIC_KEY_CHORDS] = { 7, 7, 5, 9, 2, 4 };
static int basic_key_suff[BASIC_KEY_CHORDS] = { 0, 2, 0, 1, 1, 1 };

char *get_basic_key_chord(int key, int idx)
{
	static char buf[16];

	*buf = '\0';
	strcat(buf, chordnames[(key + basic_key_chords[idx]) % 12]);
	if (basic_key_suff[idx] != 0)
		strcat(buf, chordsuff[basic_key_suff[idx]]);
	return buf;
}

void generate_basic_key_chords(int key, int idx)
{
	generate_chords((key + basic_key_chords[idx]) % 12, basic_key_suff[idx]);
}

#if 0
static void draw_line(int rel_pos)
{
	int i;

	for (i = 0; i < BAR_RANGE; i++)
		printf("|-%c-", rel_pos == i ? '*' : '-');
	printf("|\n");
}

static void disp_chords(void)
{
	int i, j;
	int start_bar, pos;

	for (i = 0; i < chords_cnt; i++) {
		start_bar = get_chord_start_bar(i);
		for (j = 0; j < STRINGS_CNT; j++) {
			pos = get_pos(chords[i], j);
			printf("%d", start_bar + 1);
			draw_line(pos - start_bar - 1);
		}
		printf("\n");
	}
}

int main(void)
{
	generate_chords(0, 0);
	disp_chords();
	return 0;
}
#endif
