/*
 *  Copyright (c) 2007 Jiri Benc <jbenc@upir.cz>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <glib.h>
#include <glib/gprintf.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#ifdef OS2006
#include <hildon-widgets/hildon-program.h>
#else
#include <hildon/hildon-program.h>
#endif
#include <libosso.h>
#include <gconf/gconf.h>
#include <gconf/gconf-client.h>
#include "prefs.h"
#include "chordlib.h"

/*
 * C  C# D  D# E  F  F# G  G# A  Bb B/H
 * 0  1  2  3  4  5  6  7  8  9  10 11
 */

static signed char tuning[][4] = {
	{ 9, 4, 0, 7 },
	{ 11, 6, 2, 9 },
	{ 4, 11, 7, 2 },
	{ 4, 9, 2, 7 },
	{ 9, 2, 7, 0 },
	{ 9, 2, 7, 0 },
	{ 4, 11, 7, 2 },
};
#define TUNING_CNT	((int)(sizeof(tuning) / sizeof(tuning[0])))

static char *tuning_name[TUNING_CNT] = {
	"Ukulele (AECG)",
	"Classical ukulele (BF#DA)",
	"Baritone ukulele (EBGD)",
	"Mandolin (EADG)",
	"Tenor mandola (ADGC)",
	"Tenor banjo (ADGC)",
	"Tenor banjo - Chicago tuning (EBGD)",
};

static int tuning_gconf_map[TUNING_CNT] =
	{ 0, 3, 4, 1, 5, 6, 7, };
	/* 2 has to be missing to preserve custom tunning from 0.3 */

#define GCONF_PREFIX	"/apps/maemo/s4chords"

static GConfClient *conf;
static void (*changed_callback)(void);

static GtkComboBox *box_lang, *box_tun;
static GtkComboBox *(box_strings[STRINGS_CNT]);

static char *(action_names[A_CNT]) = {
	"Nothing",
	"Fullscreen",
	"Quit",
	"Previous chord variants",
	"Next chord variants",
	"Previous key",
	"Next key",
	"First chord in the key",
	"Previous chord in the key",
	"Next chord in the key",
	"Reset type of the first chord",
	"Previous type of the first chord",
	"Next type of the first chord",
};

struct key_info {
	unsigned map;
	int act;
	int def;
	char *name;
};

static struct key_info keys[] = {
	{ GDK_Page_Up,	0, A_PREV_SCR,		"Tap left"	},
	{ GDK_Page_Down,0, A_NEXT_SCR,		"Tap right"	},
	{ GDK_Left,	0, A_PREV_SUBCHORD,	"Left"		},
	{ GDK_Right,	0, A_NEXT_SUBCHORD,	"Right"		},
	{ GDK_Up,	0, A_PREV_CHORD,	"Up"		},
	{ GDK_Down,	0, A_NEXT_CHORD,	"Down"		},
	{ GDK_Return,	0, A_FIRST_SUBCHORD,	"Select"	},
	{ GDK_F8,	0, A_PREV_SUFF,		"Zoom -"	},
	{ GDK_F7,	0, A_NEXT_SUFF,		"Zoom +"	},
	{ GDK_F6,	0, A_FULLSCREEN,	"Fullscreen"	},
	{ GDK_Escape,	0, A_FIRST_SUFF,	"Back"		},
};
#define KEY_CNT		((int)(sizeof(keys) / sizeof(keys[0])))

static GtkComboBox *(box_keys[KEY_CNT]);

static void set_cust_tuning(int tun, unsigned cust_tun)
{
	if (tun >= 0 && tun < TUNING_CNT)
		set_tuning(tuning[tun]);
	else {
		signed char cust[4];
		int i;

		for (i = 0; i < STRINGS_CNT; i++) {
			cust[i] = cust_tun & 0xf;
			if (cust[i] >= 12)
				cust[i] = 0;
			cust_tun >>= 4;
		}
		set_tuning(cust);
	}
}

static int tun_map_to_gconf(int tun)
{
	if (tun < 0 || tun >= TUNING_CNT)
		return 0xffff;
	return tuning_gconf_map[tun];
}

static int tun_map_from_gconf(int tun)
{
	int i;

	for (i = 0; i < TUNING_CNT; i++)
		if (tuning_gconf_map[i] == tun)
			return i;
	return 0xffff;
}

static int conf_get_int(const gchar *key, int deflt)
{
	GConfValue *gval;
	int res;

	if (!conf)
		return deflt;
	/* Yes, I know, I should use schema for default values, etc. etc.,
	 * and have the code dozen times longer as the result with no
	 * benefit for the user. Have I already said I consider this gconf
	 * shit totally brain-damaged idiotic piece of crap? */
	gval = gconf_client_get_without_default(conf, key, NULL);
	if (!gval)
		return deflt;
	if (gval->type == GCONF_VALUE_INT)
		res = gconf_value_get_int(gval);
	else
		res = deflt;
	gconf_value_free(gval);
	return res;
}

static void conf_set_int(const gchar *key, int val)
{
	GConfValue *gval;

	if (!conf)
		return;
	gval = gconf_value_new(GCONF_VALUE_INT);
	gconf_value_set_int(gval, val);
	gconf_client_set(conf, key, gval, NULL);
	gconf_value_free(gval);
}

static int conf_get_key_action(int idx, int deflt)
{
	char buf[64];

	g_snprintf(buf, 64, GCONF_PREFIX "/key%02d", idx);
	return conf_get_int(buf, deflt);
}

static void conf_set_key_action(int idx, int val)
{
	char buf[64];

	g_snprintf(buf, 64, GCONF_PREFIX "/key%02d", idx);
	conf_set_int(buf, val);
}

void prefs_init(void (*changed)(void))
{
	int i;
	int lang, tun;
	unsigned cust_tun;

	changed_callback = changed;

	conf = gconf_client_get_default();
	if (conf)
		gconf_client_add_dir(conf, GCONF_PREFIX, GCONF_CLIENT_PRELOAD_ONELEVEL, NULL);

	lang = conf_get_int(GCONF_PREFIX "/chords_lang", LANG_EN);
	tun = tun_map_from_gconf(conf_get_int(GCONF_PREFIX "/tuning", 0));
	cust_tun = conf_get_int(GCONF_PREFIX "/custom_tuning", 0x5420);
	for (i = 0; i < KEY_CNT; i++) {
		keys[i].act = conf_get_key_action(i, keys[i].def);
		if (keys[i].act < 0 || keys[i].act >= A_CNT)
			keys[i].act = A_NOTHING;
	}

	set_chord_lang(lang);
	set_cust_tuning(tun, cust_tun);
}

static void box_tun_changed(GtkComboBox *box, gpointer data)
{
	int i;
	gboolean sens;

	(void)data;

	sens = gtk_combo_box_get_active(box) == TUNING_CNT;
	for (i = 0; i < STRINGS_CNT; i++)
		gtk_widget_set_sensitive(GTK_WIDGET(box_strings[i]), sens);
}

static GtkWidget *tune_layout(void)
{
	GtkTable *tab;
	int cnt, i, j;
	unsigned tun;

	tab = GTK_TABLE(gtk_table_new(3, 5, FALSE));

	box_lang = GTK_COMBO_BOX(gtk_combo_box_new_text());
	for (i = 0; i < LANG_CNT; i++)
		gtk_combo_box_append_text(box_lang, langnames[i]);
	i = conf_get_int(GCONF_PREFIX "/chords_lang", LANG_EN);
	if (i < 0 || i >= LANG_CNT)
		i = 0;
	gtk_combo_box_set_active(box_lang, i);
	
	gtk_table_attach(tab, gtk_label_new("Chords type"), 0, 1, 0, 1, 0, 0, 0, 4);
	gtk_table_attach(tab, GTK_WIDGET(box_lang), 1, 5, 0, 1, GTK_EXPAND | GTK_FILL, 0, 0, 4);

	box_tun = GTK_COMBO_BOX(gtk_combo_box_new_text());
	for (i = 0; i < TUNING_CNT; i++)
		gtk_combo_box_append_text(box_tun, tuning_name[i]);
	gtk_combo_box_append_text(box_tun, "Custom");
	tun = tun_map_from_gconf(conf_get_int(GCONF_PREFIX "/tuning", 0));
	if (tun >= TUNING_CNT)
		tun = TUNING_CNT;
	gtk_combo_box_set_active(box_tun, tun);
	g_signal_connect(G_OBJECT(box_tun), "changed", G_CALLBACK(box_tun_changed), NULL);
	gtk_table_attach(tab, gtk_label_new("Tuning"), 0, 1, 1, 2, 0, 0, 0, 4);
	gtk_table_attach(tab, GTK_WIDGET(box_tun), 1, 5, 1, 2, GTK_EXPAND | GTK_FILL, 0, 0, 4);

	tun = conf_get_int(GCONF_PREFIX "/custom_tuning", 0x5420);
	cnt = get_chordnames_cnt();
	for (i = 0; i < STRINGS_CNT; i++) {
		box_strings[i] = GTK_COMBO_BOX(gtk_combo_box_new_text());
		gtk_widget_set_size_request(GTK_WIDGET(box_strings[i]), 120, -1);
		for (j = 0; j < cnt; j++)
			gtk_combo_box_append_text(box_strings[i], chordnames[j]);
		gtk_combo_box_set_active(box_strings[i], tun & 0xf);
		tun >>= 4;
		gtk_table_attach(tab, GTK_WIDGET(box_strings[i]), i + 1, i + 2, 2, 3, GTK_EXPAND | GTK_FILL, 0, 0, 0);
	}
	gtk_table_attach(tab, gtk_label_new("Custom tuning"), 0, 1, 2, 3, 0, 0, 0, 4);

	box_tun_changed(box_tun, NULL);

	gtk_table_set_col_spacing(tab, 0, 8);
	return GTK_WIDGET(tab);
}

static GtkComboBox *keys_combo(void)
{
	GtkComboBox *box;
	int i;

	box = GTK_COMBO_BOX(gtk_combo_box_new_text());
	for (i = 0; i < A_CNT; i++)
		gtk_combo_box_append_text(box, action_names[i]);
	return box;
}

static GtkWidget *keys_layout(void)
{
	GtkScrolledWindow *win;
	GtkTable *tab;
	int i;

	win = GTK_SCROLLED_WINDOW(gtk_scrolled_window_new(NULL, NULL));
	gtk_scrolled_window_set_policy(win, GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
	tab = GTK_TABLE(gtk_table_new(KEY_CNT, 2, FALSE));

	for (i = 0; i < KEY_CNT; i++) {
		box_keys[i] = keys_combo();
		gtk_combo_box_set_active(box_keys[i], keys[i].act);
		gtk_table_attach_defaults(tab, gtk_label_new(keys[i].name), 0, 1, i, i + 1);
		gtk_table_attach_defaults(tab, GTK_WIDGET(box_keys[i]), 1, 2, i, i + 1);
	}

	gtk_table_set_col_spacing(tab, 0, 8);

	gtk_scrolled_window_add_with_viewport(win, GTK_WIDGET(tab));
	gtk_widget_set_size_request(GTK_WIDGET(win), -1, 400);
	return GTK_WIDGET(win);
}

static GtkWidget *main_layout(void)
{
	GtkNotebook *nb;

	nb = GTK_NOTEBOOK(gtk_notebook_new());
	gtk_notebook_append_page(nb, tune_layout(), gtk_label_new("Tuning"));
	gtk_notebook_append_page(nb, keys_layout(), gtk_label_new("Keys"));
	gtk_notebook_set_scrollable(nb, FALSE);
	gtk_widget_show_all(GTK_WIDGET(nb));
	return GTK_WIDGET(nb);
}

void run_prefs(GtkWindow *parent)
{
	GtkDialog *win;
	int res, i;
	int lang, tun;
	unsigned cust_tun;

	win = GTK_DIALOG(gtk_dialog_new_with_buttons("Preferences", parent,
		GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_NO_SEPARATOR,
		GTK_STOCK_OK, GTK_RESPONSE_ACCEPT,
		GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT,
		NULL));
	if (!win)
		return;
	gtk_container_add(GTK_CONTAINER(win->vbox), main_layout());
	res = gtk_dialog_run(win);
	if (res == GTK_RESPONSE_ACCEPT) {
		lang = gtk_combo_box_get_active(box_lang);
		tun = gtk_combo_box_get_active(box_tun);
		cust_tun = 0;
		for (i = STRINGS_CNT; i >= 0; i--) {
			cust_tun <<= 4;
			cust_tun |= gtk_combo_box_get_active(box_strings[i]);
		}
		conf_set_int(GCONF_PREFIX "/chords_lang", lang);
		conf_set_int(GCONF_PREFIX "/tuning", tun_map_to_gconf(tun));
		conf_set_int(GCONF_PREFIX "/custom_tuning", cust_tun);
		set_chord_lang(lang);
		set_cust_tuning(tun, cust_tun);

		for (i = 0; i < KEY_CNT; i++) {
			keys[i].act = gtk_combo_box_get_active(box_keys[i]);
			conf_set_key_action(i, keys[i].act);
		}

		changed_callback();
		/* Yes, this is exactly what gconf manual says shouldn't be
		 * done. I don't care. Gconf design is so f*cked up that
		 * this is the only way how to deal with it and remain sane.
		 * If I hook up callbacks to all of these keys, I would end
		 * up with dozen redraws of the screen after confirming the
		 * preferences dialog box. */
	}
	gtk_widget_destroy(GTK_WIDGET(win));
}

int get_key_action(unsigned keyval)
{
	int i;

	for (i = 0; i < KEY_CNT; i++) {
		if (keys[i].map == keyval) 
			return keys[i].act;
	}
	return -1;
}
